#!/usr/bin/perl
#


use strict;
use Cwd  qw(abs_path);
use File::Basename;
use File::Glob ':glob';
use Digest::MD5 qw(md5_hex);

my $EXECDIR = abs_path(dirname($0)); $EXECDIR =~ /^(.*)$/; $EXECDIR = $1;

require("$EXECDIR/cfg.pl");
my %CONFIG = %{get_config()};

my $LOGNAME = 'error.log.txt';

my $PROCESSIDFILE = '/tmp/medeller_job_pid_'.md5_hex($EXECDIR).'.txt';


## MAKE SURE ONLY ONE INSTANCE OF THIS SCRIPT IS RUNNING AT A TIME
#
if (-f $PROCESSIDFILE)
{
  open(my $pidfile, $PROCESSIDFILE);
  my $procid = <$pidfile>;
  close($pidfile);
  chomp $procid;
  if (process_exists($procid))
  {
    exit 0;
  }
}
{
  my $procid = $$;
  open(my $pidfile, ">$PROCESSIDFILE");
  print $pidfile $procid;
  close($pidfile);
}
#
##


my @alljobs;
my $sometimeago;
my $numjobs;


while (1) # infinite loop
{
  @alljobs = bsd_glob("$CONFIG{'JOBSDIR'}/[0-9]*_*", GLOB_MARK); # stuff that begins with a number and contains an underscore
  $sometimeago = time() - (48 * 3600); # 48 hours ago
  $numjobs = 0;
  for my $JOB (@alljobs)
  {
    if ($JOB =~ m%^(.*)/$%) # name ends with a slash --> it's a directory
    {
      $JOB = $1;
      my @status = stat($JOB);
      my $mtime = $status[9];
      if ($mtime < $sometimeago) # job is old
      {
        job_delete($JOB); # might not work, if web server runs as a different user than this script
      }
      else
      {
        if (job_is($JOB, 'queued'))
        {
          if (!job_is($JOB, 'running'))
          {
            job_run($JOB);
          }
          #~ # Only allow a limited number of jobs to run in parallel
          #~ #
          #~ if (!job_is($JOB, 'done') and !job_is($JOB, 'crashed') and (!$CONFIG{'PARALLELJOBS'} or (++$numjobs >= $CONFIG{'PARALLELJOBS'})))
          #~ {
            #~ exit 0;
          #~ }
        }
      }
    }
  }

  sleep 10; # Sleep for a few seconds before we check for more jobs
}

exit 0;
# End of execution. This line should never be reached, since we're running the above code in an infinite loop.


#####################################


# Check status of a job
sub job_is
{
  my ($JOB, $stat) = @_;
  if (-f "$JOB/status.$stat")
  {
    return 1;
  }
  my $OUTDIR = $JOB;
  $OUTDIR =~ s/^$CONFIG{'JOBSDIR'}/$CONFIG{'CACHEDIR'}/;
  return -f "$OUTDIR/status.$stat";
}


#####################################


# Delete a job
sub job_delete
{
  my ($dir) = @_;
  system("rm -rf $dir 2>/dev/null");
  $dir =~ s/^$CONFIG{'JOBSDIR'}/$CONFIG{'CACHEDIR'}/;
  system("rm -rf $dir 2>/dev/null");
}


# Run a job
sub job_run
{
  my ($JOB) = @_;
  $JOB =~ s%/$%%;
  
  my $OUTPUTDIR = $JOB;
  $OUTPUTDIR =~ s/^$CONFIG{'JOBSDIR'}/$CONFIG{'CACHEDIR'}/;
  

  my $result_id = $JOB;
  $result_id =~ s/^$CONFIG{'JOBSDIR'}\///;
  if ($result_id !~ /^\w+$/)
  {
    die "Illegal result_id: $result_id";
  }
  my $LOGFILE   = "$OUTPUTDIR/$LOGNAME";
  

  # Create output directory
  #
  system("mkdir -m0755 -p $OUTPUTDIR");
  unless (-d $OUTPUTDIR)
  {
    die "Could not create output directory: $OUTPUTDIR";
  }
  chdir($OUTPUTDIR) or die;
  
  
  umask(0022);
  
  
  set_status($OUTPUTDIR, 'running');


  my $options = '';
  open(my $optsfile, "$JOB/$CONFIG{'OPTIONSFILE'}") or die;
  while (my $line = <$optsfile>)
  {
    if ($line !~ /^\s*$|^\s*#/)
    {
      chomp $line;
      $options .= "$line ";
    }
  }
  close($optsfile);
  
  my $command = "$options 1>>$LOGFILE 2>&1";
  print "$command\n";
  system($command);
  
  if ($? == -1)
  {
    finish($OUTPUTDIR, "Could not execute the job!\nchild ($result_id) failed to execute: $!\n");
    return;
  }
  elsif ($? & 127)
  {
    finish($OUTPUTDIR, "Could not execute job!\n".sprintf("child ($result_id) died with signal %d, %s coredump\n",
          ($? & 127),  ($? & 128) ? 'with' : 'without'));
    return;
  }
  else
  {
    my $val = $? >> 8;
    if ($val != 0)
    {
      open(my $logf, "<", "$LOGFILE");
      local $/ = undef;
      my $logtxt = <$logf>;
      close($logf);
      
      finish($OUTPUTDIR, "Job '$result_id' failed (code ".sprintf("%d", $val)."). See the log file below:\n\n$logtxt\n");
      return;
    }
  }

  #~ # Ugly fix for Apache's dislike of PDB files
  #~ #
  #~ foreach my $file (glob("$OUTPUTDIR/*.pdb"))
  #~ {
    #~ $file =~ m/^(.*?)(\.pdb)$/;
    #~ $file = $1.$2; # untaint $file
    #~ 
    #~ my $newfile = $1.'.atm'; # replace .pdb extension with .atm
    #~ 
    #~ system("mv $file $newfile");
  #~ }
  
  system("chmod 0755 $OUTPUTDIR/*");
  system("chmod 0664 $LOGFILE");

  set_status($OUTPUTDIR, 'done');
}
  

#########################################################################


sub set_status
{
  my ($jobpath, $statusname, $message) = @_;
  unless(defined($message))
  {
    $message = 'OK';
  }
  write_file("$jobpath/status.$statusname", $message);
}


sub finish
{
  my ($OUTDIR, $message) = @_;
  if (!defined($message))
  {
    $message = '(no crash message available)';
  }
  
  warn $message;

  set_status($OUTDIR, 'crashed', $message);

  return 1;
}

sub process_exists
{
  my ($pid) = @_;
  return kill(0, $pid);
}
