#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# MEDELLER : template-based membrane protein structure prediction
#  ==> corebuilder : building the common core between the target and template structure(s)
#
# Author: Sebastian Kelm
# Created: 10/09/2009
#
#



import sys, os.path
from array import array
from collections import deque

from prosci.common import *
from prosci.util.ali import Ali
from prosci.util.pdb import residueCode
from prosci.util.residue import ResidueList, is_residue_consecutive, is_residue_numbering_consecutive
from prosci.util.pdb3d import compare_structures, rmsd_backbone, dist
from prosci.util.gaps import isGap, gappify, deGappify, getGapChars, length_ungapped, GappySequenceError
from prosci.loops.fread import meld



def map_struc_to_seq(model, seq, renumber=False):
  if isinstance(model, ResidueList):
    model_residues = model
  else:
    model_residues = ResidueList(model)
  
  modelseq = model_residues.get_seq()
  assert len(modelseq) == len(model_residues)
  
  model_fragments = [ResidueList([model_residues[0]])]
  for i in xrange(1,len(model_residues)):
    if is_residue_consecutive(model_residues[i-1], model_residues[i]) and is_residue_numbering_consecutive(model_residues[i-1], model_residues[i]):
      model_fragments[-1].append(model_residues[i])
    else:
      model_fragments.append(ResidueList([model_residues[i]]))
  
  seq = seq.upper()
  #modelseq_aligned = ""
  indeces = []
  i = 0
  for frag in model_fragments:
    fragseq = frag.get_seq()
    j = seq.find(fragseq, i)
    if j < 0:
      raise Exception("Cannot map structure to sequence:\n%s\n%s\n%s"%(modelseq, seq, str([f.get_seq() for f in model_fragments])))
    indeces.extend(range(j, j+len(fragseq)))
    #modelseq_aligned += "-" * (j-i) + fragseq
    i = j + len(fragseq)
    
    if renumber:
      # Renumber residues to show their position in the native sequence
      # (starting with residue 1)
      #
      for k, residue in enumerate(frag):
        for atom in residue:
          atom.ires = 1 + j + k
          atom.inscode = ""
  
  #modelseq_aligned += "-" * (len(seq) - len(modelseq_aligned))
  
  return indeces



def calculateAccuracy(mymodel, target_pdb, target_residue_indeces, model_pdb, normalise_by=None):
    scores, scores_model = None, None
    
    if not target_pdb:
      return scores, scores_model
    
    if isinstance(mymodel, ResidueList):
      mymodel = mymodel.to_pdb()
    if isinstance(target_pdb, ResidueList):
      target_pdb = target_pdb.to_pdb()
    if isinstance(model_pdb, ResidueList):
      model_pdb = model_pdb.to_pdb()
    
    target_subset     = target_pdb.get_residues(target_residue_indeces)
    target_subset_seq = target_subset.get_seq()
    mymodel_seq       = mymodel.get_seq()
    
    # DEBUG
    if len(target_subset_seq) != len(mymodel_seq):
      print target_subset_seq
      print mymodel_seq
    
    if normalise_by is None:
      normalise_by=target_pdb.rescount()
    
    scores = compare_structures(target_subset, mymodel, target_subset_seq, mymodel_seq, realign=True, modify=False, normalise_by=normalise_by)
    
    if model_pdb:
      model_subset = model_pdb.get_residues(target_residue_indeces)
      scores_model = compare_structures(target_subset, model_subset, target_subset.get_seq(), model_subset.get_seq(), realign=True, modify=False, normalise_by=target_pdb.rescount())
    
    return scores, scores_model


def combineSelection(sel1, sel2, function=OR):
    "change sel1 to reflect the combination of sel1 and sel2 using the boolean operator 'function'. Valid functions are: AND, OR, NAND, NOR, XOR"
    assert len(sel1) == len(sel2)
    for i in xrange(len(sel1)):
      sel1[i] = function(sel1[i], sel2[i])



class SubstTable:
    def __init__(self, data):
      self.data=[] # list of arrays of floats
      self.collabels=[] # list of strings
      self.rowlabels=[] # list of strings
      self.parse_file(data)
    
    def parse_file(self, fname):
      f = open(fname)
      try:
        lines = f.readlines()
      finally:
        f.close()
      self.collabels = lines[0].split()
      if self.collabels[0] in ("", '""', "''"):
        self.collabels.pop(0)
      
      for i in xrange(1, len(lines)):
        numbers = array('f')
        fields = lines[i].split()
        for i in xrange(1, len(fields)):
          numbers.append(float(fields[i]))
        self.rowlabels.append(fields[0])
        self.data.append(numbers)
    
    def get(self, col, row):
      x = self.collabels.index(col)
      y = self.rowlabels.index(row)
      return self.data[x][y]



class SubstTableSelector:
    def __init__(self, directory, prefix=""):
      
      # backwards compatibility
      if directory and prefix and isinstance(directory, dict) and (isinstance(prefix, list) or isinstance(prefix, tuple)):
        value2filename = directory
        envtitles = prefix
        self.data = dict()
        self.envs = envtitles
        for k in value2filename:
          self.data[k] = SubstTable(value2filename[k])
        return
      
      if not prefix and not os.path.isdir(directory):
        prefix = os.path.basename(directory)
        directory = os.path.dirname(directory)
      
      fileprefix = "%s/%sENV_"%(directory, prefix)
      filesuffix = "_LOG_ODDS.mat"
      
      factors = []
      f = file("%s/%sdef"%(directory, prefix))
      for line in f:
          line = line.strip()
          if line and not line.startswith("#"):
            factors.append(line.split(';'))
            assert len(factors[-1]) == 3
            assert len(factors[-1][1]) == len(factors[-1][2])
      f.close()
      
      factors.sort()
      
      self.envs = [x[0] for x in factors]
      values = deque([x[1] for x in factors])
      fnames = deque([x[2] for x in factors])
      
      value_combinations = [x for x in values[0]]
      fname_combinations = [x for x in fnames[0]]
      
      for i in xrange(1, len(self.envs)):
        while len(value_combinations[0]) <= i:
          for j in xrange(len(values[i])):
            value_combinations.append(value_combinations[0]+values[i][j])
            fname_combinations.append(fname_combinations[0]+fnames[i][j])
          del value_combinations[0]
          del fname_combinations[0]
          
      
      self.values2name = {}
      for v, f in zip(value_combinations, fname_combinations):
        self.values2name[v] = f
          
      
      self.data = {}
      for i in xrange(len(value_combinations)):
        val = value_combinations[i]
        fnm = fname_combinations[i].replace('?', '_')
        self.data[val] = SubstTable(file(fileprefix+fnm+filesuffix))
    
    def __getitem__(self, valuestring):
      return self.data[valuestring]
    
    def getEnvTitles(self):
      return self.envs



class CoordinateCopier:
    
    def __init__(self, target_entrygroup, template_entrygroup, template_pdb=None, template_directory=None, template_number=0, chaincode = 'A'):
        assert len(target_entrygroup)
        assert len(template_entrygroup)
        assert isinstance(target_entrygroup, Ali.EntryGroup)
        assert isinstance(template_entrygroup, Ali.EntryGroup)
        
        self.chaincode = chaincode
        
        if template_pdb is None:
          template_pdb = ResidueList(template_entrygroup.getMasterEntry().getStructureFilename(template_directory))
        else:
          template_pdb = ResidueList(template_pdb)
        
        self.windowLeft  = 1
        self.windowRight = 1
        
        self.target_group     = target_entrygroup
        self.target_id        = target_entrygroup.getCode()
        self.target_seq       = target_entrygroup.getMasterEntry().get_seq()
        self.target_mapping   = None # list of integers, mapping from column indeces to target residue indeces
        
        self.template_number  = template_number
        self.template_group   = template_entrygroup
        self.template_id      = template_entrygroup.getCode()
        self.template_seq     = template_entrygroup.getMasterEntry().get_seq()
        #self.template_pdb     = template_pdb.get_atoms_by_type(("N", "CA", "C", "O", "CB"))
        self.template_pdb     = template_pdb
        self.template_mapping = None # list of integers, mapping from column indeces to template residue indeces
        
        self.selection        = None # list of booleans, with length equal to the alignment length; True marks a column as selected
        self.mask             = None # as self.selection; True marks a column as masked, i.e. never to be selected
        self.bfactors         = None # list of numbers that signify how trustworthy the coordinates are
        
        # These are only set when calling setSubstTables(.) and calculateSubstitutionScores()
        #
        self.substitution_tables = None # SubstTableSelector object
        self.substitution_scores = None # list of scores (may include None)
        
        # Sanity checks
        #
        assert len(self.target_seq)
        assert len(self.template_seq)
        assert len(self.target_seq) == len(self.template_seq)
        assert len(self.template_pdb)
        assert self.template_group.hasMasterEntry()
        
        
        # Map alignment position to residue index in TARGET
        #
        self.target_mapping = []
        n=0
        for c in self.target_seq:
          if isGap(c):
            self.target_mapping.append(-1)
          else:
            self.target_mapping.append(n)
            n+=1
        
        #assert deGappify(self.template_pdb.get_seq()) == deGappify(self.template_seq), "Template sequences are not equal in alignment and structure files! id: %s"%(self.template_id)
        
        # Map alignment position to residue index in TEMPLATE
        #
#         self.template_mapping = []
#         n=0
#         for c in self.template_seq:
#           if isGap(c):
#             self.template_mapping.append(-1)
#           else:
#             self.template_mapping.append(n)
#             n+=1
#         
#         assert len(self.template_mapping) == len(self.target_seq)
#         
        
        # Map template structure to its own sequence in the alignment
        # (in case the alignment contains the full template sequence but
        #  the structure is missing some residues)
        #
        template_struc2seq = map_struc_to_seq(self.template_pdb, deGappify(self.template_seq))
        
        # Make a cumulative gap count at each position of the ungapped template sequence
        gapcounts = []
        n = 0
        for i, c in enumerate(self.template_seq):
          if isGap(c):
            n += 1
          else:
            gapcounts.append(n)
        
        # Create blank mapping of template sequence to structure
        self.template_mapping = [-1] * len(self.target_seq)
        
        # Reverse the structure-to-sequence mapping to fill the sequence-to-structure 
        # mapping, while accounting for the gaps in the template sequence within the input
        # alignment
        #
        for i, v in enumerate(template_struc2seq):
          v = template_struc2seq[i] = v + gapcounts[v]
          self.template_mapping[v] = i
        
        
        
        # Initialise selection
        #
        self.selection = [False] * len(self.template_mapping)
        self.bfactors  = [None]   * len(self.template_mapping)
        
        # Initialise selection mask
        #
        self.resetMask()
        
    
    def getSelectedTargetResidueIndeces(self):
        "return the list of residue indeces in the target that are covered by the template, given the current selection"
        selected_residues = []
        for i,v in enumerate(self.selection):
          if v and not isGap(self.target_seq[i]) and not isGap(self.template_seq[i]):
            selected_residues.append(self.target_mapping[i])
        return selected_residues
    
    
    def getSelectedCoordinates(self):
        "return the coordinates (Pdb object) of the selected residues of the template structure"
        
        coords = ResidueList([])
        coords.code = self.target_id
        
        fragMap, fragments = self.getSelectedFragments()
        
        for i, frag in enumerate(fragments):
          coords.extend(frag.getCoordinates())
        
        coords.renumber_atoms()
        
        
        return coords
    
    
    def selectNone(self):
        "Make sure nothing is selected"
        for i in xrange(len(self.selection)):
          self.selection[i] = False
          self.bfactors[i]  = None
  
    def selectAll(self, bfactor=None):
        "Select all unmasked alignment columns"
        for i,v in enumerate(self.mask):
          self.selection[i] = not v
          self.bfactors[i]  = bfactor
    
    
    def combineSelection(self, sel, function=OR, bfactor=None):
        "Combine the current selection with a given selection using a boolean operator. Valid operators are: AND, OR, NAND, NOR, XOR"
        oldsel=self.selection[:]
        combineSelection(self.selection, sel, function)
        self.applySelectionMask()
        for i,v in enumerate(self.selection):
          if v and not oldsel[i]:
            self.bfactors[i] = bfactor
    
    
    def clearMask(self):
      "Makes sure only alignment gaps are masked."
      self.mask = [False] * len(self.selection)
      self.maskGaps(widenGaps=False)
    
    def resetMask(self):
      "Makes sure only alignment gaps and adjacent residues are masked."
      self.mask = [x < 0 for x in self.template_mapping]
      self.maskGaps()
    
    
    def maskGaps(self, widenGaps=True):
      "Masks alignment gaps and adjacent residues, in addition to any existing mask."
      values = getGapChars()
      
      mask=[False]*len(self.mask)
      for seq in (self.target_seq, self.template_seq):
        for v in values:
          sel = findAllBoolean(seq, v)
          combineSelection(mask, sel, OR)
      
      if widenGaps:
        # Widen any masked regions by 1 residue on each side (if gap is not in both seqs)
        # Part 1
        #
        prev=0
        for i in xrange(1,len(mask)):
          if self.canSkipColumn(i):
            continue
          if mask[i]:
            mask[prev]=True
          prev=i
        
        # Widen any masked regions by 1 residue on each side (if gap is not in both seqs)
        # Part 2
        #
        prev=len(mask)-1
        for i in rxrange(0,len(mask)-1):
          if self.canSkipColumn(i):
            continue
          if mask[i]:
            mask[prev]=True
          prev=i
      
      combineSelection(self.mask, mask, OR)

    
    
    def maskGappyLoops(self, function=OR):
      "Masks loops with alignment gaps, in addition to any existing mask."
      self.maskLoops(function, maskGappy=True, maskLong=False)
    
    
    def maskLongLoops(self, function=OR, maxLoopLength=3):
      "Masks loops longer than a given length cut-off, in addition to any existing mask."
      self.maskLoops(function, maskGappy=False, maskLong=True, maxLoopLength=maxLoopLength)
    
    
    def maskLoops(self, function, maskGappy, maskLong, maxLoopLength=3):
      "Internal function. Masks loops that meet certain criteria, in addition to any existing mask."
      assert maxLoopLength >= 0
      assert len(self.mask) == len(self.target_seq)
      assert len(self.template_seq) == len(self.target_seq)
      
      entrytype = "secondary structure and phi angle"
      
      seq = self.template_group[entrytype].get_seq()
      mask = [False] * len(seq)
      
      # Find and mask loops containing or adjacent to a gap.
      # Gaps present in both sequences are ignored
      #
      loopStart=-1
      loopIsGappy=False
      loopLength=0
      for i,v in enumerate(seq):
        gapTemp = isGap(v)
        gapTarg = isGap(self.target_seq[i])
        
        if self.canSkipColumn(i) or (gapTemp and gapTarg):
          continue # gap in both template and target, so ignore this alignment column
        
        if loopStart >= 0:                                        # we are inside a loop
          if gapTarg:                                             # target has a gap
            loopIsGappy = True
          
          if v == "C":                                            # next template residue is still inside a loop
            loopLength+=1
          elif gapTemp:                                           # template has a gap
            loopIsGappy = True
          else:                                                   # end of loop
            #if loopIsGappy or loopLength > maxLoopLength:
            if (maskLong  and loopLength > maxLoopLength) or \
               (maskGappy and loopIsGappy):
                  for j in xrange(loopStart, i):
                    mask[j] = True                                # mask loop, including any gaps inside or next to it
            loopStart = -1
            loopIsGappy = False
            loopLength = 0
        else:                                                     # we are not inside a loop yet
          if v == "C":                                            # start of loop
            loopStart = i
            loopLength = 1
            if i==0 or (isGap(seq[i-1]) or isGap(self.target_seq[i-1])): #  (N-terminal loops are always masked)
              loopIsGappy=True
      
      # Finish off loops at end of sequence (C-terminal loops are always masked)
      #
      if loopStart >= 0: # and (loopIsGappy or loopLength > maxLoopLength):
        for j in xrange(loopStart, len(seq)):
          mask[j] = True                                          # mask loop, including any gaps inside or next to it
        
        
      combineSelection(self.mask, mask, function)
    
    
    def selectByValue(self, entrytype, values, function=OR, wordcutoffs=0, minwordlength=1, bfactor=None):
      "Selects all residues with a certain annotation, minus any masked residues."
      assert wordcutoffs >= 0
      assert minwordlength > 0
      
      if None == function:
        self.selectNone()
        function = OR
      
      newselection = [False] * len(self.selection)
      #print self.template_id
      for v in values:
        sel = findAllBoolean(self.template_group[entrytype].get_seq(), v)
        #print "".join([str(int(v)) for v in sel])
        combineSelection(newselection, sel, OR)
        #print self.selectionToString(), "*"
      
      
      #print self.selectionToString(), "*"
      
      words, gaplist = tokenize([False], newselection)
      reduceTokens(words, gaplist, wordcutoffs, minwordlength)
      newselection = deTokenize(False, words, gaplist)
      
      
      self.combineSelection(newselection, function, bfactor)
      
      
      return self.selection.count(True)
      
      #print self.selectionToString(), "*"
      #print
    
    
    def setSubstTables(self, ESSTs):
      self.substitution_tables=ESSTs
    
    
    def calculateSubstitutionScores(self):
      "Calculates the raw substitution scores for each alignment column."
      if not self.substitution_tables:
        raise IllegalStateError("Must call setSubstTables(.) before calling this method")
      
      envs = self.substitution_tables.getEnvTitles() # list of strings
      
      substitution_scores = [None]*len(self.selection) # list of numbers
      
      for i,c1 in enumerate(self.target_seq):
        if self.template_mapping[i] < 0:
           continue
        
        if isGap(c1):
          # Target has a gap - cannot calculate substitution score for this column
          continue
        
        c2 = self.template_seq[i]
        if isGap(c2):
          # Template has a gap - cannot calculate substitution score for this column
          continue
          
        env = ""
        for etitle in envs:
          env += self.template_group[etitle].get_seq()[i]
        substitution_scores[i] = self.substitution_tables[env].get(c2, c1)
      
      self.substitution_scores = substitution_scores
    
    
    def setDummySubstitutionScores(self, value=1):
      "Sets dummy substitution scores for each alignment column."
      
      substitution_scores = [None]*len(self.selection) # list of numbers
      
      for i,c1 in enumerate(self.target_seq):
        if self.template_mapping[i] < 0:
           continue
        
        if isGap(c1):
          # Target has a gap - cannot calculate substitution score for this column
          continue
        
        c2 = self.template_seq[i]
        if isGap(c2):
          # Template has a gap - cannot calculate substitution score for this column
          continue
          
        substitution_scores[i] = value
      
      self.substitution_scores = substitution_scores
    
    
    def applySelectionMask(self):
      #print self.maskToString(), "m"
      # Remove masked residues from selection
      #
      for i in xrange(len(self.selection)):
        if self.mask[i]:
          self.selection[i] = False
          self.bfactors[i]  = None
    
    
    def canSkipColumn(self, column):
      gapTarg = isGap(self.target_seq[column])
      gapTemp = isGap(self.template_seq[column])
      return gapTarg and gapTemp
    
    
    def deselect(self, index):
      self.selection[index] = False
      self.bfactors[index]  = None
    
    def select(self, index, bfactor):
      self.selection[index] = True
      self.bfactors[index]  = bfactor
    
    
    def setWindowSize(self, windowLeft, windowRight):
      self.windowLeft = windowLeft
      self.windowRight = windowRight
    
    
    def getSmoothedScore(self, cand):
      if not self.substitution_scores:
        self.calculateSubstitutionScores()
      
      # Determine correct LEFT window limit
      #
      i=cand-1                                # cursor
      l=cand-self.windowLeft                       # initial left window limit
      while i >= l and l>=0:                  # we're walking left, from our starting residue to the left window limit
        if self.mask[i]:            # we hit a masked column. can we skip this or do we cut off the window here?
          if self.canSkipColumn(i):
            l-=1                              # skippable gap column - shift window limit one further away and continue
          else:
            l=i+1                             # gap or masked residue - cut off the window and stop
            break
        i-=1                                  # walk one residue further towards the left window limit
      
      # Determine correct RIGHT window limit
      #
      i=cand+1                                # cursor
      r=cand+1+self.windowRight                      # initial right-most residue inside the window
      while i < r and r<=len(self.substitution_scores):         # we're walking right, from our starting residue to the right window limit
        if self.mask[i]:            # we hit a masked column. can we skip this or do we cut off the window here?
          if self.canSkipColumn(i):
            r+=1                              # skippable gap column - shift window limit one further away and continue
          else:
            r=i                               # gap or masked residue - cut off the window and stop
            break
        i+=1                                  # walk one residue further towards the right window limit
        
      return average(self.substitution_scores[max(0,l):r], ignoreNone=True)
    
    
    def getFragmentScore(self, cand):
      scRight=scLeft=self.getSmoothedScore(cand)
      
      i=cand+1
      while i < len(self.substitution_scores):
        if self.canSkipColumn(i):
          pass
        elif self.selection[i] and self.substitution_scores[i] != None:
          scRight+=self.getSmoothedScore(i)
        else:
          break
        i+=1
      
      j=cand-1
      while j >= 0:
        if self.canSkipColumn(j):
          pass
        elif self.selection[j] and self.substitution_scores[j] != None:
          scLeft+=self.getSmoothedScore(j)
        else:
          break
        j-=1
      
      if i == cand+1:
        return scLeft
      
      if j == cand-1:
        return scRight
      
      return max(scLeft, scRight)
    
    
    def getLeftIndex(self, i):
      j=1
      while True:
        if i-j < 0:
          return None
        if not self.canSkipColumn(i-j):
          return i-j
        j+=1
    
    def getRightIndex(self, i):
      j=1
      while True:
        if i+j >= len(self.selection):
          return None
        if not self.canSkipColumn(i+j):
          return i+j
        j+=1
    
    
    def extendSelectionOrdered(self, useFragmentScore=True, bfactor=None):
      """extendSelectionOrdered(useFragmentScore=True) : Extend selection fragments by one residue
        
        returns (score, index) if the selection was altered, (None, None) otherwise
      """
      if not self.substitution_scores:
        self.calculateSubstitutionScores()
      
      
      assert len(self.substitution_scores) == len(self.selection)
      assert len(self.substitution_scores) == len(self.mask)
      
      
      # Get list of non-selected residues adjacent to a selection (including gaps)
      #
      candidates = []
      if len(self.selection)>1 and not self.selection[0] and self.selection[1]:
        candidates.append(0)
      for i in xrange(1, len(self.selection)-1):
        if not self.selection[i] and (self.selection[i+1] or self.selection[i-1]):
          candidates.append(i)
      if len(self.selection)>1 and not self.selection[-1] and self.selection[-2]:
        candidates.append(len(self.selection)-1)
      
      
      # Remove masked residues from candidates
      # This takes care of any gaps.
      # Gaps that appear in both sequences are skipped and candidates modified or deleted accordingly.
      #
      i=0
      while i<len(candidates):
        if self.mask[candidates[i]]: # residue is masked. must not select.
            # Is there a gap in both sequences?
            #
            if not self.canSkipColumn(candidates[i]): # no
              del candidates[i]
            else: # yes
              if candidates[i]==0 or candidates[i] == len(self.selection)-1:
                del candidates[i] # we've reached one end of the alignment and cannot extend further
              
              else:
                # Check if there is a selected column to our right
                #
                j_right = self.getRightIndex(candidates[i])
                j_left = self.getLeftIndex(candidates[i])
                goingLeft = (j_right is not None) and self.selection[j_right]
                goingRight = (j_left is not None) and self.selection[j_left]
                
                if goingLeft:
                  # We are extending towards the left
                  
                  # Try next residue to the left
                  candidates[i] = j_left
                  if (candidates[i] is None) or (i>0 and candidates[i] <= candidates[i-1]) or self.selection[candidates[i]]:
                    del candidates[i]
                  #else:
                    #print candidates[i]
                elif goingRight:
                  # We are extending towards the right
                  
                  # Try next residue to the right
                  candidates[i] = j_right
                  if (candidates[i] is None) or (i+1<len(candidates) and candidates[i] >= candidates[i+1]) or self.selection[candidates[i]]:
                    del candidates[i]
                  #else:
                    #print candidates[i]
                else:
                  del candidates[i]
        else:
          i+=1
      del i
      
      
      
      if not candidates:
        return (None, None)
      
      
      if useFragmentScore:
        getScore = self.getFragmentScore
      else:
        getScore = self.getSmoothedScore
      
      
      bestCandidate = 0
      bestScore = getScore(candidates[0])
      for i in xrange(1,len(candidates)):
        sc = getScore(candidates[i])
        if sc > bestScore:
          bestCandidate = i
          bestScore = sc
      
      assert not self.selection[candidates[bestCandidate]], "BUG: selecting already selected residue: %d" %(candidates[bestCandidate])
      
      self.select(candidates[bestCandidate], bfactor)
      
      return (bestScore, candidates[bestCandidate])
    
    
    def extendThroughDisulphides(self, function=OR, wordextension=1, minwordlength=3, bfactor=None):
      "Selects disulphides and surrounding residues, if one of the bonded cysteins is already selected, taking into account any masked residues."
      assert wordextension >= 0
      assert minwordlength > 0
      entrytype = "disulphide"
      values = "T"
      
      if None == function:
        self.selectNone()
        function = OR
      
      newselection = [False] * len(self.selection)
      for v in values:
        sel = findAllBoolean(self.template_group[entrytype].get_seq(), v)
        #print "".join([str(int(v)) for v in sel])
        combineSelection(newselection, sel, OR)
        #print self.selectionToString(), "*"
      cysteins = []
      for i, x in enumerate(newselection):
        if x:
          cysteins.append((i, self.template_pdb[self.template_mapping[i]].CB.copy()))
      
      candidates = []
      for i1, a1 in cysteins:
        if self.selection[i1] and not self.mask[i1]:
          for i2, a2 in cysteins:
            if not self.selection[i2] and not self.mask[i2]:
              d = dist(a1, a2)
              if d > 3 and d < 5:
                candidates.append(i2)
      
      for i in candidates:
        newselection = [False] * len(self.selection)
        newselection[i] = True
        wordlen = 1
        for x in xrange(wordextension):
          j = self.getRightIndex(i)
          if j is None or self.mask[j]:
            break
          newselection[j] = True
          wordlen += 1
        for x in xrange(wordextension):
          j = self.getLeftIndex(i)
          if j is None or self.mask[j]:
            break
          newselection[j] = True
          wordlen += 1
        
        if wordlen >= minwordlength:
          self.combineSelection(newselection, function, bfactor)
      
      return self.selection.count(True)
    
    
    def selectionToString(self):
      return "".join([str(int(v)) for v in self.selection])
    
    def maskToString(self):
      return "".join([str(int(v)) for v in self.mask])
    
    
    
    def fractionOfTemplateSelected(self):
      seq = self.template_seq
      
      assert len(seq) == len(self.selection)
      
      selected=0
      total   =0
      for i in xrange(len(seq)):
        if isGap(seq[i]):
          continue
        total += 1
        if self.selection[i]:
          selected+=1
      
      return float(selected)/float(total)
    

    def fractionOfTargetSelected(self):
      seq = self.target_seq
      
      assert len(seq) == len(self.selection)
      
      selected=0
      total   =0
      for i in xrange(len(seq)):
        if isGap(seq[i]):
          continue
        total += 1
        if self.selection[i]:
          selected+=1
      
      return float(selected)/float(total)
    
    
    def fractionOfCoreSelected(self):
      seq = self.target_seq
      
      assert len(seq) == len(self.selection)
      
      # Length of N-terminal gap
      gapN=0
      for i in xrange(len(seq)):
        if isGap(seq[i]):
          continue # skip gaps in target
        if not self.selection[i]: # residue not covered by selected core
          gapN+=1
        else:
          break
      
      # Length of C-terminal gap
      gapC=0
      for i in rxrange(len(seq)):
        if isGap(seq[i]):
          continue # skip gaps in target
        if not self.selection[i]: # residue not covered by selected core
          gapC+=1
        else:
          break
      
      selected=0
      total   =0
      for i in xrange(len(seq)):
        if isGap(seq[i]):
          continue # skip gaps in target
        total += 1
        if self.selection[i]: # residue covered by selected core
          selected+=1
      
      return float(selected)/float(total-gapN-gapC)
    
    
    def getSelectedFragments(self):
        fragList = []
        
        sel  = self.selection
        fragMap = [-1]*len(sel)
        
        inFrag=False
        fragNumber=-1
        for i,v in enumerate(sel):
          if self.canSkipColumn(i):
            continue
          if sel[i]:
            if not inFrag:
              fragNumber+=1
              inFrag=True
            fragMap[i]=fragNumber
          else:
            inFrag=False
        
        for i,v in enumerate(fragMap):
          if v>=0:
            if v >= len(fragList):
              fragList.append(Fragment(self, v))
            fragList[v].addResidue(i)
        
        return fragMap, fragList
    
    
    def getMaxFragmentScore(self):
        fragMap, fragList = self.getSelectedFragments()
        maxScore = float("-inf")
        for frag in fragList:
          maxScore = max(maxScore, frag.getScore())
        return maxScore


class Fragment:
    def __init__(self, parent, fragnum, isMerged=False):
      self.parent  = parent
      self.fragnum = fragnum
      self.indeces = []
      self.coords  = ResidueList([]) # coords for each residue separately
      self.scores  = []
      
      if isMerged:
        self.template_number = -1
        self.template_id = ""
      else:
        self.template_number = self.parent.template_number
        self.template_id     = self.parent.template_id
    
    def __cmp__(self, other):
      if type(None) == type(other):
        return 1
      v=cmp(self.first(), other.first())
      if v==0:
        v=cmp(self.last(), other.last())
      return v
    
    def getTemplateId(self):
      return self.template_id
    def getTemplateNumber(self):
      return self.template_number
    def getFragmentNumber(self):
      return self.fragnum
    
    def getTargetId(self):
      return self.parent.target_id
    
    def first(self):
      return self.indeces[0]
    
    def last(self):
      return self.indeces[-1]
    
    def addResidue(self, i, residue=None, score=None):
      self.indeces.append(i)
      
      if residue is not None:
        self.coords.append(residue.copy())
      else:
        residue = self.parent.template_pdb[self.parent.template_mapping[i]].copy()
        self.coords.append(residue)
        
        # Relabel the residue's coordinates
        #
        residue.set_type(residueCode(self.parent.target_seq[i]))
        ires = 1+self.parent.target_mapping[i]
        for atm in residue:
          atm.ires = ires
          atm.inscode = ""
          atm.chain = self.parent.chaincode
          if self.parent.bfactors[i] is not None:
            atm.b = float(max(0,self.parent.bfactors[i]))
      
      if score!=None:
        self.scores.append(score)
      else:
        self.scores.append(self.parent.getSmoothedScore(i))
    
    
    def getUngappedLength(self):
      return len(self.indeces)
    def getGappedLength(self):
      return 1 + (self.last() - self.first())
    
    def getScore(self):
      return sum(self.scores)
    
    
    def getCoordinates(self):
      coords=self.coords.deep_copy()
      coords.code = self.getTargetId()
      #coords.renumber_atoms()
      return coords
    
    
    def getIndeces(self):
      return self.indeces
    def getSelectedTargetResidueIndeces(self):
      return [self.parent.target_mapping[i] for i in self.indeces]
    def getSelectedTemplateResidueIndeces(self):
      return [self.parent.template_mapping[i] for i in self.indeces]
    
    
    def overlaps(self, other):
      return self.getIndeces() and other.getIndeces() and not ( other.first() > self.last() or other.last() < self.first() )
    
    
    def includes(self, other):
      return self.getIndeces() and other.getIndeces() and other.first() >= self.first() and other.last() <= self.last()
    
    
    def getGappedOverlapIndeces(self, other):
      if not self.overlaps(other):
        raise ValueError("Fragments do not overlap")
      
      if self.includes(other):
        return other.first(), other.last()
      if other.includes(self):
        return self.first(), self.last()
      if self.first() <= other.last() and self.first() >= other.first():
        return self.first(), other.last()
      if other.first() <= self.last() and other.first() >= self.first():
        return other.first(), self.last()
      
      raise RuntimeError("BUG: Overlap not cought by conditions")
    
    
    def getOverlapFraction(self, other):
      ls = self.getUngappedLength()
      lo = other.getUngappedLength()
      loverlap = len(set(self.getIndeces()) & set(other.getIndeces()))
      #return float(loverlap) / (ls+lo-loverlap)
      return float(loverlap) / min(ls, lo)
    
    
    def getOverlap(self, other):
      first, last = self.getGappedOverlapIndeces(other)
      
      a = Fragment(self.parent, self.getFragmentNumber())
      b = Fragment(other.parent, other.getFragmentNumber())
      
      indeces = sorted(set(self.getIndeces()) & set(other.getIndeces()))
      
      # Make sure the start and end of the overlap are correct
      #
      if not (indeces and indeces[0] == first and indeces[-1] == last):
        raise GappySequenceError("Fragment overlap is gappy %d-%d: %s; (%d.%d x %d.%d)"%(first, last, str(indeces), self.getTemplateNumber(), self.getFragmentNumber(), other.getTemplateNumber(), other.getFragmentNumber()))
      
      # Make sure there are no gaps in the middle of the overlap.
      # Having a gap in both sequences at the same place is fine.
      #
      prev=indeces[0]
      for i in indeces:
        if i>prev+1:
          raise GappySequenceError("Fragment overlap is gappy")
        prev=i
      
      iself  = self.indeces.index(first)
      iother = other.indeces.index(first)
      
      for x,i in enumerate(indeces):
        a.addResidue(i, self.coords[iself+x], self.scores[iself+x])
        b.addResidue(i, other.coords[iother+x], other.scores[iother+x])
      
      return a, b
    
    
    def rmsd(self, other):
      fragA, fragB = self.getOverlap(other)
      
      coordA = fragA.getCoordinates()
      coordB = fragB.getCoordinates()
      
      return (rmsd_backbone(coordA, coordB), fragA, fragB, coordA, coordB)
    
    
    def merge(self, other, fragNum, maxRmsd=1.0, maxOverlap=0.5):
      # Determine left and right fragments
      #
      if self.first() < other.first():
        a=self
        b=other
      else:
        a=other
        b=self
      
      if a.includes(b) or b.includes(a):
        raise ValueError("One fragment includes the other. Should not merge!")
      
      if a.getOverlapFraction(b) > maxOverlap:
        return None
      
      try:
        rmsd, overlapA, overlapB, coordA, coordB = a.rmsd(b)
      except GappySequenceError:
        # Fragment overlap is gappy, i.e. not perfectly aligned. Do not merge.
        return None
      
      if rmsd > maxRmsd:
        return None
      
      #averageCoordinates(coordA, coordB) # saves result coordinates in coordA
      meld(coordA, coordB) # saves result coordinates in coordA
      
      overlapCoords = coordA
      assert len(overlapCoords) == overlapA.getUngappedLength()
      
      
      # Start building merged fragment
      #
      merged = Fragment(a.parent, fragNum, isMerged=True)
      
      # Add non-overlapping part of fragment a
      for x,i in enumerate(a.indeces):
        if i >= overlapA.first():
          break
        merged.addResidue(i, a.coords[x], a.scores[x])
      
      # Add overlapping, averaged, coordinates
      # Also merge score vectors for the overlap
      totWeight = float(1+len(overlapA.indeces))
      for x,i in enumerate(overlapA.indeces):
        W=float(x+1) # score weight
        merged.addResidue(i, overlapCoords[x], ((totWeight-W)*overlapA.scores[x] + W*overlapB.scores[x])/totWeight)
      
      # Add non-overlapping part of fragment b
      for x,i in enumerate(b.indeces):
        if i > overlapB.last():
          merged.addResidue(i, b.coords[x], b.scores[x])
      
      return merged


class CoreBuilder:
    def __init__(self, target_entrygroup, template_entrygroups, template_pdbs=None, template_directory=None, verbose=False, score_in_occupancy=False, useFragmentScore=True, chaincode='A'):
      if template_pdbs==None:
        template_pdbs = [None]*len(template_entrygroups)
      else:
        assert len(template_pdbs)==len(template_entrygroups)
      
      self.verbose      = verbose # print lots of messages?
      self.score_in_occupancy = score_in_occupancy # Should we write substitution scores to the occupancy column?
      
      self.maskLoops    = False # call setSubstTables() to set this
      self.model        = None  # call setBenchmark() to set this
      self.native       = None  # call setBenchmark() to set this
      self.useFragmentScore = useFragmentScore
      
      self.cc           = []    # list of CoordinateCopier objects, one per template
      self.scoreHistory = None  # list of HistoryDataPoint objects, describing the core building process
      
      self.fragmentMaps  = []   # selected residues mapped to their fragment number, one per template, 
      self.fragmentLists = []   # lists of Fragment objects, one list per template
      self.fragmentScores= []   # lists of fragment scores, one list per template
      
      self.merged        = False # list of merged fragments
      
      self.chosen        = []   # list of individual fragments to be included in the 3d model.
                                # each entry is of the format: [template_number, fragment_number, fragment_score, Fragment]
      
      
      self.setWindowSize() # sets the default window sizes
      
      for i,eg in enumerate(template_entrygroups):
        self.cc.append(CoordinateCopier(target_entrygroup, template_entrygroups[i], template_pdbs[i], template_directory, template_number=i, chaincode=chaincode))
    
    
    def setVerbose(self, v):
      self.verbose = v
    
    
    def setWindowSize(self, windowLeft=1, windowRight=1):
      for copier in self.cc:
        copier.setWindowSize(windowLeft, windowRight)
    
    
    def setSubstTables(self, ESSTs, maskLoops=True):
      self.maskLoops = maskLoops
      for copier in self.cc:
        copier.setSubstTables(ESSTs)
    
    
    def setBenchmark(self, native, model):
      assert model==None or native.get_seq() == model.get_seq(), "Model and Native target PDB sequences differ. Perhaps missing residues in native structure?"
      self.native = native
      self.model  = model
    
    
    def getTargetId(self):
      return self.cc[0].target_id
    def getTargetGroup(self):
      return self.cc[0].target_group
    def getTargetSeq(self):
      return self.cc[0].target_seq
    
    
    class HistoryDataPoint:
      def __init__(self, phase, coverage, coreCoverage, score, index, strucscores_native=None, strucscores_model=None):
        self.phase = phase
        self.coverage = coverage
        self.coreCoverage = coreCoverage
        self.score = score
        self.index = index
        self.strucscores_native = strucscores_native
        self.strucscores_model  = strucscores_model
      
      def copy(self):
        return CoreBuilder.HistoryDataPoint(self.phase, self.coverage, self.coreCoverage, self.score, self.index, self.strucscores_native, self.strucscores_model)
      
      def __iter__(self):
        yield self.phase
        yield self.coverage
        yield self.coreCoverage
        yield self.score
        yield self.index
        if self.strucscores_native:
          for sc in sorted(self.strucscores_native):
            yield self.strucscores_native[sc]
        if self.strucscores_model:
          for sc in sorted(self.strucscores_model):
            yield self.strucscores_model[sc]
      
      def __str__(self):
        return join("\t", tuple(self))+"\n"
      
      def __repr__(self):
        return "CoreBuilder.HistoryDataPoint(%s)"%(join(",", tuple(self)))
    # end class HistoryDataPoint


    def collectResult(self, templatenum, phase, score, index):
      results = self.scoreHistory[templatenum]
      copier  = self.cc[templatenum]
      
      coverage  = copier.fractionOfTargetSelected()
      coreCoverage = copier.fractionOfCoreSelected()
      
      if coverage <= 0.0:
        return False
      
      strucscores, strucscores_model = None, None
      if self.native:
        coreModel = copier.getSelectedCoordinates()
        target_residue_indeces = copier.getSelectedTargetResidueIndeces() # residues covered by any template
        assert target_residue_indeces, "BUG: coverage>0 but no residues selected??"
        strucscores, strucscores_model = calculateAccuracy(coreModel, self.native, target_residue_indeces, self.model)
        results.append(CoreBuilder.HistoryDataPoint(phase, coverage, coreCoverage, score, index, strucscores, strucscores_model))
      else:
        results.append(CoreBuilder.HistoryDataPoint(phase, coverage, coreCoverage, score, index))
      return True


    def selectMinimalCore(self, wordcutoffs=2, minwordlength=2, annotation_type="membrane layer", annotation_values="T"):
      """Select the minimal core to copy from the template"""
      
      emptyCores = 0
      self.scoreHistory = []
      
      for templatenum, copier in enumerate(self.cc):
        #print copier.maskToString(), "m"
        if self.maskLoops:
          copier.maskGappyLoops()
          #print copier.maskToString(), "m"
          copier.maskLongLoops(maxLoopLength=3)
          #print copier.maskToString(), "m"
        numSel = copier.selectByValue(annotation_type, annotation_values, OR, wordcutoffs, minwordlength, bfactor=0)
        
        if self.verbose:
          print "#Minimal (end cutoff = %d, min word length = %d): %s" % (wordcutoffs, minwordlength, copier.selectionToString())
        
        if numSel == 0:
          emptyCores += 1
        
        self.scoreHistory.append([])
        self.collectResult(templatenum, 0, copier.getMaxFragmentScore(), -1)
      
      if emptyCores >= len(self.cc):
        raise ValueError("ERROR: None of the %d templates have a transmembrane core. Target or templates not transmembrane proteins?\n"%len(self.cc))
      elif emptyCores > 0:
        sys.stderr.write("WARNING: %d of %d templates have no transmembrane core. Target or templates not transmembrane proteins? Using remaining templates.\n"%(emptyCores, len(self.cc)))
    
    
    def selectAll(self):
      """Select all aligned coordinates (the naive model) as the core."""
      emptyCores = 0
      self.scoreHistory = []
      for templatenum, copier in enumerate(self.cc):
        copier.clearMask() # nothing is masked
        numSel = copier.selectAll(bfactor=0)
        
        if numSel == 0:
          emptyCores += 1
        
        self.scoreHistory.append([])
        self.collectResult(templatenum, 0, copier.getMaxFragmentScore(), -1)
        
      if emptyCores >= len(self.cc):
        raise ValueError("ERROR: None of the %d templates are aligned to the target!\n"%len(self.cc))
      elif emptyCores > 0:
        sys.stderr.write("WARNING: %d of %d templates are not aligned to the target.\n"%(emptyCores, len(self.cc)))
    
    
    def extendThroughDisulphides(self, wordextension=1, minwordlength=3):
      """Gradually extend model through disulphides"""
      for c in self.cc:
        c.extendThroughDisulphides(wordextension=wordextension, minwordlength=minwordlength, bfactor=3)
    
    
    def extendCore(self, phases=(1,2,3)):
      """Gradually extend model from the minimal core and, if in BENCHMARK mode, test accuracy all along the way"""
      
      phases = sorted(set(phases) & set([1,2,3]))
      
      for templatenum, copier in enumerate(self.cc):
        copier.resetMask()
        if self.maskLoops:
          copier.maskGappyLoops()
          copier.maskLongLoops(maxLoopLength=3)
        
        results=self.scoreHistory[templatenum]
        
        score=results[-1].score
        index=results[-1].index
        
        if score == float("-inf"):
          # No minimal core selected
          return
        
        #~ if not self.collectResult(templatenum, 0, score, index): # collect result for phase 0 (minimal core)
          #~ return
        
        keepGoing=True
        for phase in phases:
          
          if not keepGoing:
            break
          
          ## Calculate cut-off score for loop residue addition
          #
          #
          prevPhase             = results[-1].phase
          prevPhaseFinalScore   = results[-1].score
          prevPhaseAverageScore = 0.0
          n=1
          while n<len(results)-1:
            if results[-n].phase != prevPhase:
              break
            prevPhaseAverageScore += results[-n].score
            n+=1
          n-=1
          if n > 0:
            prevPhaseAverageScore /= float(n)
          del n
          
          cutoffScore = float((2.0*prevPhaseAverageScore + 1.0*prevPhaseFinalScore) / 3.0)
          #
          #
          ##
          
          if not self.maskLoops and phase > 1:
              # We weren't masking loops in the first place,
              # so no point trying to unmask them now.
              # Stop after phase 1.
              #
              break
          
          if phase == 1:
            pass
          elif phase == 2:
            # Unmask all loops, except for the gappy ones, and resume copying
            copier.resetMask()
            copier.maskGappyLoops()
          elif phase == 3:
            # Unmask all loops, even the gappy ones, and resume copying
            copier.resetMask()
          
          i=0
          while keepGoing:
            # Extract the selected coordinates from the template
            #
            score1, index1 = copier.extendSelectionOrdered(useFragmentScore=self.useFragmentScore, bfactor=phase)
            if None == score1:
              # Stop this copying phase when we can't copy any more (i.e. all unmasked residues have been selected)
              break
            
            score, index = score1, index1
            
            i+=1
            if self.verbose:
              print "#Extended (stop at gaps) %3d: %s" % (i, copier.selectionToString())
            
            # Collect score & index
            #
            self.collectResult(templatenum, phase, score, index)
            
            if phase >= 2: # we have a solid core already
              #if score < 0: # our score is negative
                if score < cutoffScore: # we're below the cutoff score
                  if self.verbose:
                    print "#Fulfilled stopping conditions. Backtracking to a local score maximum. %3d: %s" % (i, copier.selectionToString())
                  while len(results) > 1 and results[-1].phase == phase and results[-2].score >= score:
                    assert index >= 0
                    assert index == results[-1].index
                    copier.deselect(index) # undo residue selection
                    results.pop() # remove datapoint
                    score = results[-1].score
                    index = results[-1].index
                    if self.verbose:
                      print "#Backtracking to a local score maximum. %3d: %s" % (i, copier.selectionToString())
                  keepGoing=False
            
          del i
        
        # end for phase in phases
      
      return self.scoreHistory
    
    
    def getNewUnalignedTargetAli(self):
        oldentry = self.getTargetGroup().getMasterEntry()
        seq = deGappify(oldentry.get_seq())
        struclign = "structure:%s:%4d :%s:%4d :%s:unknown:unknown:-1:-1"%(self.getTargetId().replace("P1;", ""), 1, "@", len(seq), "")
        return Ali( ">%s\n%s\n%s*\n"%(self.getTargetId(),struclign,seq) )
    
    
    def prepareFragments(self):
        self.fragmentMaps  = []
        self.fragmentLists = []
        self.fragmentScores= []
        
        for template in self.cc:
          fragMap, fragLst = template.getSelectedFragments()
          self.fragmentMaps.append(fragMap)
          self.fragmentLists.append(fragLst)
          self.fragmentScores.append([x.getScore() for x in fragLst])
        
        if self.verbose:
          print self.getTargetId()
          print self.getTargetSeq()
          print
          for i,m in enumerate(self.fragmentMaps):
            #print self.cc[i].selectionToString()
            #print self.cc[i].maskToString()
            x=""
            for v in m:
              if v<0:
                x+='_'
              else:
                x+=str(v)
            print self.cc[i].template_id
            print x
            print "Scores:",
            for s in self.fragmentScores[i]:
              print "%.1f, "%s,
            print
            print
    
    
    def createMergedFragments(self, maxRmsd = 1.0, maxOverlap=0.5):
        if self.merged:
          return
        if len(self.cc)<=1:
          return
        
        if not self.fragmentLists:
          self.prepareFragments()
        
        #
        # Create pairs of merged fragments
        #
        
        merged = []
        left2right = {}
        #right2left = {}
        
        for i, templ1 in enumerate(self.fragmentLists):
          for frag1 in templ1:
            for j in xrange(i+1, len(self.fragmentLists)):
              templ2 = self.fragmentLists[j]
              for frag2 in templ2:
                if frag1.overlaps(frag2) and not frag1.includes(frag2) and not frag2.includes(frag1):
                  m = frag1.merge(frag2, len(merged), maxRmsd, maxOverlap)
                  if m != None:
                    merged.append(m)
                    
                    l = (frag1.getTemplateNumber(), frag1.getFragmentNumber())
                    r = (frag2.getTemplateNumber(), frag2.getFragmentNumber())
                    
                    if frag1 > frag2:
                      temp = l
                      l = r
                      r = temp
                    
                    #m.left  = l # first constituent single fragment
                    m.prev  = l # previous to last constituent single fragment
                    m.right = r # last constituent single fragment
                    
                    if l in left2right:
                      left2right[l].append(r)
                    else:
                      left2right[l] = [r]
                    
                    #if r in right2left:
                      #right2left[r].append(l)
                    #else:
                      #right2left[r] = [l]
                    
        merged.sort()
        
        
        #
        # From the pairs of merged fragments, create all possible
        # combinations of multiple merged fragments
        #
        
        
        def id2frag(fragId):
          frag = self.fragmentLists[fragId[0]][fragId[1]]
          assert frag.getTemplateNumber() == fragId[0]
          assert frag.getFragmentNumber() == fragId[1]
          return frag
        
        istart = 0
        iend   = len(merged)
        
        while iend > istart:
          for i in xrange(istart, iend):
            frag1 = merged[i] # fragment to be extended
            if frag1.right in left2right:
              # frag1's right is someone else's left
              rightFragIds = left2right[frag1.right]
              for rightId in rightFragIds:
                
                if rightId in left2right[frag1.prev]:
                  # The fragment from the previous extension overlaps
                  # with the fragment to our right. Don't add it, to
                  # avoid averaging more than two sets of coordinates
                  # on a single residue.
                  continue
                
                frag2 = id2frag(rightId) # single fragment (has no "left" or "right" data!)
                
                m = frag1.merge(frag2)
                if m != None:
                  #m.left  = frag1.left
                  m.prev  = frag1.right
                  m.right = rightId
                  merged.append(m)
          
          istart = iend
          iend = len(merged)
        # end while
        
        
        self.merged = True
        self.fragmentLists.append(merged)
        self.fragmentScores.append([f.getScore() for f in merged])
    
    
    def chooseFragmentsHeuristic(self):
        """Fast, heuristic algorithm for choosing a good combination of fragments, but not necessarily the best"""
        
        if not self.fragmentLists:
          self.prepareFragments()
        if len(self.cc) == 1:
          self.chosen = self.fragmentLists[0]
          return
        
        ranking = []
        for i in xrange(len(self.fragmentLists)):
          for j,score in enumerate(self.fragmentScores[i]):
            ranking.append([i, j, score, self.fragmentLists[i][j]])
        
        def compareRank(b, a):
          v=cmp(a[2], b[2])
          if v!=0:
            return v
          return cmp(a[-1].getUngappedLength(), b[-1].getUngappedLength())
        
        ranking.sort(compareRank)
        
        
        if self.verbose:
          print "Fragment ranking"
          for i, rank in enumerate(ranking):
            print "%2d"%i, rank[0], rank[1], "%6.1f"%rank[2]
      
      
        self.chosen = []
        for i, rank in enumerate(ranking):
          current = rank[-1]
          overlaps = False
          for previous in self.chosen:
            if previous.overlaps(current):
              overlaps = True
              break
          if not overlaps:
            self.chosen.append(current)
      
        self.chosen.sort() # sorts by first(), last() in ascending order
        
        
        if self.verbose:
          print "Chosen fragments:"
          for i, frag in enumerate(self.chosen):
            tn = frag.getTemplateNumber()
            fn = frag.getFragmentNumber()
            print "%2d"%i, tn, fn, "%6.1f"%self.fragmentScores[tn][fn], "%3d"%frag.first(), "%3d"%frag.last()
    
    
    def chooseFragmentsExhaustive(self):
        """Slow, exhaustive algorithm for choosing the best-scoring combination of fragments"""
        
        SCOFFSET=1000 # score offset for fragment scores
        INACCURACY=0.001 # leeway for score comparisons, to account for machine imprecision
        
        
        def cmpFloat(score1, score2):
          if abs(score1-score2) < INACCURACY:
            return 0
          else:
            return cmp(score1, score2)
        
        
        if not self.fragmentLists:
          self.prepareFragments()
        if len(self.cc) == 1:
          self.chosen = self.fragmentLists[0]
          return
        
        # Create good core to start with.
        # Comparisons to this core will allow dead-end elimination to speed up the exhaustive search.
        #
        self.chooseFragmentsHeuristic()
        
        heuristicScore = 0.0
        heuristicLength = 0
        for frag in self.chosen:
          heuristicScore  += frag.getScore()
          heuristicLength += frag.getUngappedLength()
        
          
        
        
        # Calculate max scores for each alignment column. Take their cumulative sums, starting from the C-terminal end. Each score is associated with one alignment column and represents the maximum possible score that a growing core can hope to add to its own score, assuming that all available fragments are 1 residue long.
        # This can be used to calculate if a growing core has any hope of ever scoring better than the heuristic core.
        # Note: these scores are NOT offset using SCOFFSET
        #
        maxRemainingScores=[0.0]*len(self.fragmentMaps[0])
        
        for col in rxrange(len(maxRemainingScores)):
          colmax=0.0
          for tmplNum, fragMap in enumerate(self.fragmentMaps):
            fragNum = fragMap[col]
            if fragNum >= 0:
              colmax=max(colmax, self.cc[tmplNum].getSmoothedScore(col))
          maxRemainingScores[col]=colmax
        
        reverseSum=0.0
        for col in rxrange(len(maxRemainingScores)):
          sc=maxRemainingScores[col]
          maxRemainingScores[col]+=reverseSum
          reverseSum+=sc
        
        del sc, reverseSum # cleanup
        
        
        
        allFragments=[]
        for i in xrange(len(self.fragmentLists)):
          for j in xrange(len(self.fragmentLists[i])):
            allFragments.append(self.fragmentLists[i][j])
        
        allFragments.sort() # sorts by first(), last() in ascending order
        
        
        scores=[]
        for sc in self.fragmentScores:
          scores.append([s+SCOFFSET for s in sc])
        
        
        def couldScoreWell(nextCol, combScore, combLen):
          """Tests if a combination of fragments with score combScore is good enough to have a hope of beating the heuristic core. nextCol is the next free column in the core"""
          
          #print combScore, combLen, SCOFFSET, nextCol, maxRemainingScores[nextCol], heuristicScore
          
          # We need to correct the combination score using SCOFFSET
          combScore -= combLen*SCOFFSET
          
          if nextCol >= len(maxRemainingScores):
            # Finished building core. Is it good enough as it is?
            return cmpFloat(combScore, heuristicScore) >= 0
          else:
            # Not finished building. Is there hope?
            maxRestScore = maxRemainingScores[nextCol]
            return cmpFloat(combScore, heuristicScore - maxRestScore) >= 0
        
        
        combinations=[[]]
        combScores=[0.0]
        #combLengths=[0]
        for frag in allFragments:
          if not combinations:
            break
          j=0
          while j < len(combinations):
            comb=combinations[j]
            if len(comb) and frag.overlaps(comb[-1]):
              j+=1
              continue
            
            
            ########################
            # Dead end elimination #
            ########################
            
            if not couldScoreWell(frag.first(), combScores[j], len(comb)):
              # There is no hope for the base combination. Not with any fragment. Delete it and stop trying to extend it.
              del combinations[j]
              del combScores[j]
              #del combLengths[j]
            else:
              # There is hope for this base combination in general.
              # How about if we extend it with the current fragment?
              #
              combSc = combScores[j]+scores[frag.getTemplateNumber()][frag.getFragmentNumber()]
              #combLen = combLengths[j]+frag.getUngappedLength()
              if couldScoreWell(frag.last()+1, combSc, len(comb)+1):
                comb2 = comb[:]
                comb2.append(frag)
                combinations.append(comb2)
                combScores.append(combSc)
                #combLengths.append(combLen)
              j+=1
          # end while
        # end for
        del j # cleanup
        
        assert len(combinations) == len(combScores)
        #assert len(combinations) == len(combLengths)
        
        
        # Remove empty combination
        #
        if len(combinations) and len(combinations[0]) == 0:
          del combinations[0]
          del combScores[0]
          #del combLengths[0]
        
        
        # Eliminate combinations that are worse than the heuristic
        #
        j=0
        while j < len(combinations):
          comb    = combinations[j]
          combSc  = combScores[j] - len(comb)*SCOFFSET
          
          if cmpFloat(combSc, heuristicScore) <  0:
            if self.verbose:
              print "Deleting combination (%.3f < %.3f):"%(combSc, heuristicScore), ["%d.%d"%(x.getTemplateNumber(), x.getFragmentNumber()) for x in comb]
            del combinations[j]
            del combScores[j]
            #del combLengths[j]
          else:
            j += 1
        del j
        
        
        if not combinations:
          if self.verbose:
            print "No core is better than the heuristic core."
          return
        
        
        # Pick the best-scoring combination
        #
        def combinationLength(comb):
          l = 0
          for f in comb:
            l += f.getUngappedLength()
          return l
        
        best=combinations[0]
        bestScore=combScores[0]
        bestLength=combinationLength(best)
        
        for i in xrange(1,len(combinations)):
          comb = combinations[i]
          sc   = combScores[i]
          l    = combinationLength(comb)
          if sc > bestScore or (sc == bestScore and l > bestLength):
            best       = comb
            bestScore  = sc
            bestLength = l
        
        bestScore -= len(best)*SCOFFSET
        
        realScore = 0.0
        for frag in best:
          realScore += frag.getScore()
        
        #assert realScore == bestScore, "Real score (%.3f) != best score (%.3f)"%(realScore, bestScore)
        
        
        if True or self.verbose:
          if best == self.chosen:
            print "Found same core using heuristic and exhaustive method"
          else:
            print "Found different core from heuristic:"
          
          print "Exhaustive (%.3f):"%bestScore, ["%d.%d"%(x.getTemplateNumber(), x.getFragmentNumber()) for x in best]
          print "Heuristic  (%.3f): "%heuristicScore, ["%d.%d"%(x.getTemplateNumber(), x.getFragmentNumber()) for x in self.chosen]
        
        self.chosen = best
    
    
    def getSelectedTargetResidueIndeces(self):
        a=set([])
        for frag in self.chosen:
          for i in frag.getSelectedTargetResidueIndeces():
            a.add(i)
        return sorted(a)
    
    
    def getSelectedTemplateResidueIndeces(self):
        a=set([])
        for frag in self.chosen:
          for i in frag.getSelectedTemplateResidueIndeces():
            a.add(i)
        return sorted(a)
    
    
    def getSelectedCoordinates(self):
        "return the coordinates (Pdb object) of the selected residues of the template structure"
        
        coords = ResidueList([])
        coords.code = self.getTargetId()
        
        for i, frag in enumerate(self.chosen):
          fragcoords = frag.getCoordinates()
          if self.score_in_occupancy:
            for j, res in enumerate(fragcoords):
              for atm in res:
                atm.occup = frag.scores[j] # Save residue score in occupancy column
          coords.extend(fragcoords)
        
        coords.renumber_atoms()
        #for res in coords:
        #  for a in res:
        #    a.occup = 1.0
        
        return coords
    
    
    def countFragments(self):
        tot=0
        for templateFrags in self.fragmentLists:
          tot += len(templateFrags)
        return tot
    
    
    def buildCoreModel(self):
        if len(self.cc) == 0:
          raise ValueError("No templates to model from.")
        
        if not self.chosen:
          self.chooseFragmentsHeuristic()
        
        if self.verbose:
          print "%d templates, %d fragments"%(len(self.cc), self.countFragments())
        
        coreModel = self.getSelectedCoordinates()
        target_residue_indeces = self.getSelectedTargetResidueIndeces()
        return coreModel, target_residue_indeces
    
